import abc
from typing import List, Dict
import time
from spruned.application import exceptions
from spruned.application.logging_factory import Logger


class RPCAPIService(metaclass=abc.ABCMeta):
    errors_ttl = 5
    max_errors_before_downtime = 1
    errors = []
    client = None
    throttling_error_codes = []

    @abc.abstractmethod
    def getrawtransaction(self, txid, **kwargs):
        pass  # pragma: no cover

    def _increase_errors(self):
        now = int(time.time())
        self.errors.append(now)

    @property
    def available(self):
        now = int(time.time())
        _errors = []
        for error in self.errors:
            if error > now - self.errors_ttl:
                _errors.append(error)
        self.errors = _errors
        return bool(len(self.errors) < self.max_errors_before_downtime)

    async def get(self, path):
        try:
            return await self.client.get(path)
        except exceptions.HTTPClientException as e:
            from aiohttp import ClientResponseError
            cause = e.__cause__
            if isinstance(cause, ClientResponseError):
                if cause.code in self.throttling_error_codes:
                    Logger.third_party.warning('throttling %s' % self.__class__.__name__)
                else:
                    Logger.third_party.exception('Error on %s: %s' % (self.__class__.__name__, e.__cause__))
                self._increase_errors()


class HeadersRepository(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_best_header(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_header_at_height(self, blockheight: int):
        pass  # pragma: no cover

    @abc.abstractmethod
    def save_header(self, blockhash: str, blockheight: int, headerbytes: bytes, prev_block_hash: str):
        pass  # pragma: no cover

    @abc.abstractmethod
    def remove_headers_after_height(self, blockheight: int):
        pass  # pragma: no cover

    @abc.abstractmethod
    def save_headers(self, headers: List[Dict]):
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_headers_since_height(self, height: int):
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_block_hash(self, height: int):
        pass  # pragma: no cover

    @abc.abstractmethod
    def remove_header_at_height(self, blockheight: int):
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_headers(self, *blockhashes: str):
        pass  # pragma: no cover



import abc
import asyncio
import time
from typing import Dict, List
from spruned.application.tools import async_delayed_task
from spruned.daemon.abstracts import ConnectionAbstract


class BaseConnection(ConnectionAbstract, metaclass=abc.ABCMeta):
    def __init__(
            self, hostname: str, use_tor=False, loop=None,
            start_score=10, timeout=10, expire_errors_after=180,
            is_online_checker: callable=None, delayer=async_delayed_task
    ):
        self._is_online_checker = is_online_checker
        self._hostname = hostname
        self._hostname = hostname
        self.use_tor = use_tor
        self._version = None
        self._on_headers_callbacks = []
        self._on_connect_callbacks = []
        self._on_disconnect_callbacks = []
        self._on_errors_callbacks = []
        self._on_peers_callbacks = []
        self.loop = loop or asyncio.get_event_loop()
        self._start_score = start_score
        self._last_header = None
        self._subscriptions = []
        self._timeout = timeout
        self._errors = []
        self._peers = []
        self._expire_errors_after = expire_errors_after
        self._is_online_checker = is_online_checker
        self.delayer = delayer

    @property
    def hostname(self):
        return self._hostname

    def add_error(self, *a):
        if len(a) and isinstance(a[0], int):
            self._errors.append(a[0])
        else:
            self._errors.append(int(time.time()))

    def is_online(self):
        if self._is_online_checker is not None:
            return self._is_online_checker()
        return True

    def add_on_header_callbacks(self, callback):
        self._on_headers_callbacks.append(callback)

    def add_on_connect_callback(self, callback):
        self._on_connect_callbacks.append(callback)

    def add_on_disconnect_callback(self, callback):
        self._on_disconnect_callbacks.append(callback)

    def add_on_peers_callback(self, callback):
        self._on_peers_callbacks.append(callback)

    def add_on_error_callback(self, callback):
        self._on_errors_callbacks.append(callback)

    async def on_header(self, header):
        self._last_header = header
        for callback in self._on_headers_callbacks:
            self.loop.create_task(callback(self))

    async def on_connect(self):
        for callback in self._on_connect_callbacks:
            self.loop.create_task(callback(self))

    async def on_error(self, error):
        if not self.is_online:
            return
        self.add_error()
        for callback in self._on_errors_callbacks:
            self.loop.create_task(callback(self, error_type=error))

    async def on_peers(self):
        for callback in self._on_peers_callbacks:
            self.loop.create_task(callback(self))

    @property
    def start_score(self):
        return self._start_score

    @property
    def version(self):
        return self._version

    @property
    def last_header(self) -> Dict:
        return self._last_header

    @property
    def subscriptions(self) -> List:
        return self._subscriptions

    @property
    def score(self):
        return self._start_score - len(self.errors)

    @property
    def errors(self):
        now = int(time.time())
        self._errors = [error for error in self._errors if now - error < self._expire_errors_after]
        return self._errors

    @property
    def peers(self):
        return self._peers



import os
import json
import asyncio

from spruned.daemon.electrod.electrod_connection import ElectrodConnectionPool
from spruned.daemon.electrod.electrod_interface import ElectrodInterface


def build(network, loop=asyncio.get_event_loop()):  # pragma: no cover
    def load_electrum_servers(network):
        _current_path = os.path.dirname(os.path.abspath(__file__))
        with open(_current_path + '/electrum_servers.json', 'r') as f:
            servers = json.load(f)
        return servers[network]

    electrod_pool = ElectrodConnectionPool(
        connections=network["electrum_concurrency"], peers=load_electrum_servers(network["electrum_servers"])
    )
    electrod_interface = ElectrodInterface(electrod_pool, loop)
    return electrod_pool, electrod_interface




"""
import asyncio
import unittest
from unittest.mock import Mock, create_autospec
import binascii
from spruned.application.spruned_vo_service import SprunedVOService
from spruned.services.thirdparty_service import ThirdPartyServiceDelegate
from test.utils import async_coro


class TestVOService(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.electrod = Mock()
        self.p2p = Mock()
        self.utxo_tracker = Mock()
        self.repository = Mock()
        self.source = create_autospec(ThirdPartyServiceDelegate)
        self.sut = SprunedVOService(
            self.electrod, self.p2p, utxo_tracker=self.utxo_tracker, repository=self.repository
        )
        hb = '000000206ad001ecab39a3267ac6db2ccea9e27907b011bc70324c00000000000000000048043a6a' \
             '574d8d826af9477804d3a4ac116a411d194c0b86d950168163c4d4232364ad5aa38955175cd60695'
        hh = '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e'
        hb = binascii.unhexlify(hb.encode())
        self.header = {
            'header_bytes': hb,
            'block_hash': hh,
            'block_height': 513979,
            'prev_block_hash': '0000000000000000004c3270bc11b00779e2a9ce2cdbc67a26a339abec01d06a'
        }
        self.response_header = {
                'bits': 391481763,
                'chainwork': 'Not Implemented Yet',
                'confirmations': 2,
                'difficulty': 'Not Implemented Yet',
                'hash': '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e',
                'height': 513979,
                'mediantime': 1521312803,
                'merkleroot': '23d4c463811650d9860b4c191d416a11aca4d3047847f96a828d4d576a3a0448',
                'nextblockhash': None,
                'nonce': 2500253276,
                'previousblockhash': '0000000000000000004c3270bc11b00779e2a9ce2cdbc67a26a339abec01d06a',
                'time': 1521312803,
                'version': 536870912,
                'versionHex': 'Not Implemented Yet'
            }

    def tearDown(self):
        self.electrod.reset_mock()
        self.utxo_tracker.reset_mock()
        self.repository.reset_mock()

    def test_getblock_non_verbose(self):
        self.repository.get_best_header.return_value = {'block_height': 513980}
        self.repository.get_block_header.return_value = self.header
        self.repository.get_block.return_value = {
                'block_hash': '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e',
                'block_bytes': binascii.unhexlify('cafebabe'.encode())
            }

        block = self.loop.run_until_complete(
            self.sut.getblock('000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e', 0)
        )
        self.assertEqual(block, 'cafebabe')

    def test_getrawtransaction(self):
        self.cache.get.return_value = None
        self.repository.get_block_header.return_value = self.header
        tx = '01000000000101531213685738c91df5ceb1537605b4e17d0e623c34ead12b9e285495cd5da9b80000000000ffffffff0248d' \
             '00500000000001976a914fa511ca56ee17f57b8190ad490c4e5bf7ef0e34b88ac951e00000000000016001458e05b9b412c3b' \
             '4f35bdb54f47376beaeb8f81aa024830450221008a6edb6ce73676d4065ffb810f3945b3c3554025d3d7545bfca7185aaff62' \
             '0cc022066e2f0640aeb0775e4b47701472b28d1018b4ab8fd688acbdcd757b75c2731b6012103dfc2e6847645ca8057120780' \
             'e5ae6fa84be76b39465cd2a5158d1fffba78b22600000000'
        self.repository.get_best_header.return_value = {'block_height': 513980}
        self.electrod.getrawtransaction.return_value = async_coro(tx)
        self.source.getrawtransaction.return_value = async_coro({
            'blockhash': '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e',
            'source': 'some unreliable api'
        })
        self.electrod.getmerkleproof.return_value = async_coro(True)
        res = self.loop.run_until_complete(
            self.sut.getrawtransaction(
                'dbae729fc6cce1bc922e66f4f12eb2b43ef57406bf5a0818eb2e73696b713b91',
                verbose=False
            )
        )
        self.assertEqual(res, tx)

    def test_getbestblockhash(self):
        self.repository.get_best_header.return_value = {'block_hash': 'cafebabe'}
        res = self.loop.run_until_complete(self.sut.getbestblockhash())
        self.assertEqual(res, 'cafebabe')

    def test_getblockhash(self):
        self.repository.get_block_hash.return_value = 'cafebabe'
        res = self.loop.run_until_complete(self.sut.getblockhash(123))
        self.assertEqual(res, 'cafebabe')

    def test_getblockheader(self):
        self.repository.get_best_header.return_value = {'block_height': 513980}
        self.repository.get_block_header.return_value = self.header
        res = self.loop.run_until_complete(
            self.sut.getblockheader(
                '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e'
            )
        )
        self.assertEqual(
            res,
            self.response_header
        )

    def test_block_count(self):
        self.repository.get_best_header.return_value = {'block_height': 6}
        res = self.loop.run_until_complete(self.sut.getblockcount())
        self.assertEqual(res, 6)

    def test_estimatefee(self):
        self.electrod.estimatefee.return_value = async_coro('fee estimation')
        res = self.loop.run_until_complete(self.sut.estimatefee(6))
        self.assertEqual(res, 'fee estimation')

    def test_getbestblockheader(self):
        self.repository.get_best_header.return_value = {
            'block_height': 513980,
            'block_hash': '0000000000000000001a0822fbaef92ef048967fa32c68f96e3d57d13183ef2b'
        }
        self.repository.get_block_header.return_value = self.header
        res = self.loop.run_until_complete(self.sut.getbestblockheader(verbose=False))
        self.assertEqual(
            res,
            '000000206ad001ecab39a3267ac6db2ccea9e27907b011bc70324c00000000000000000048043a6a'
            '574d8d826af9477804d3a4ac116a411d194c0b86d950168163c4d4232364ad5aa38955175cd60695'
        )

        res = self.loop.run_until_complete(self.sut.getbestblockheader(verbose=True))
        self.assertEqual(res, self.response_header)

    def test_getblockchaininfo(self):
        self.repository.get_best_header.return_value = self.header
        res = self.loop.run_until_complete(self.sut.getblockchaininfo())
        self.assertEqual(
            res,
            {
                'chain': 'main',
                'warning': 'spruned v0.0.1. emulating bitcoind v0.16',
                'blocks': 513979,
                'headers': 513979,
                'bestblockhash': '000000000000000000376267d342878f869cb68192ff5d73f5f1953ae83e3e1e',
                'difficulty': None,
                'chainwork': None,
                'mediantime': 1521312803,
                'verificationprogress': 0,
                'pruned': False
             }
        )

    def test_gettxout_ok(self):
        TXID = 'aa'*32
        INDEX = 0
        self.source.gettxout.return_value = async_coro(
            {
                'in_block': 123,
                'value_satoshi': 10000,
                'script_asm': 'asm',
                'script_hex': '61736d',
                'script_type': 'script_type',
                'addresses': ['1btc'],
                'unspent': True
            }
        )
        self.electrod.listunspents.return_value = async_coro([
            {'tx_hash': TXID, 'tx_pos': INDEX}
        ])
"""


import io
import binascii
import time
from typing import Dict, List
from pycoin.block import Block
from pycoin.tx.Tx import Tx
from spruned.application import utils, exceptions
from spruned.application.database import ldb_batch
from spruned.application.logging_factory import Logger
from spruned.application.tools import deserialize_header

TRANSACTION_PREFIX = b'\x00'
BLOCK_PREFIX = b'\x01'


class BlockchainRepository:
    def __init__(self, session, storage_name, dbpath):
        self.storage_name = storage_name
        self.session = session
        self.dbpath = dbpath
        self._cache = None

    def set_cache(self, cache):
        self._cache = cache

    def get_key(self, name: str, prefix=b''):
        if isinstance(prefix, str):
            prefix = prefix.encode()
        if isinstance(name, str):
            name = binascii.unhexlify(name.encode())
        return (prefix and (prefix + b'.') or b'') + name

    async def async_save_block(self, block: Dict, tracker=None, callback=None):
        res = self.save_block(block, tracker)
        callback and callback(res)

    @ldb_batch
    def save_block(self, block: Dict, tracker=None) -> Dict:
        saved = self._save_block(block)
        tracker and tracker.track(saved['key'], saved['size'])
        return saved

    @ldb_batch
    def save_blocks(self, *blocks: Dict) -> List[Dict]:
        saved = []
        for block in blocks:
            saved.append(self._save_block(block))
        return saved

    @ldb_batch
    def _save_block(self, block: Dict) -> Dict:
        _block = block['block_object']
        key = self.get_key(_block.id(), prefix=BLOCK_PREFIX)
        self.session.put(self.storage_name + b'.' + key, block['block_bytes'])
        block['key'] = key
        block['size'] = len(block['block_bytes'])
        return block

    @ldb_batch
    def save_transaction(self, transaction: Dict) -> Dict:
        blockhash = binascii.unhexlify(transaction['block_hash'].encode())
        data = transaction['transaction_bytes'] + blockhash
        key = self.get_key(transaction['txid'], prefix=TRANSACTION_PREFIX)
        self.session.put(self.storage_name + b'.' + key, data)
        return transaction

    @ldb_batch
    def save_transactions(self, *transactions: Dict) -> List[Dict]:
        saved = []
        for transaction in transactions:
            saved.append(self.save_transaction(transaction))
        return saved

    def get_block(self, blockhash: str) -> (None, Dict):
        key = self.get_key(blockhash, prefix=BLOCK_PREFIX)
        data = self.session.get(self.storage_name + b'.' + key)
        if not data:
            Logger.leveldb.debug('%s not found under key %s', blockhash, key)
            return
        header = deserialize_header(data[:80])
        return {
            'block_hash': header['hash'],
            'block_bytes': data,
            'header_bytes': data[:80],
            'timestamp': header['timestamp']
        }

    def get_transaction(self, txid) -> (None, Dict):
        return self._get_transaction(txid)

    def _get_transaction(self, txid: (str, bytes)):
        key = self.get_key(txid, prefix=TRANSACTION_PREFIX)
        data = self.session.get(self.storage_name + b'.' + key)
        if not data:
            return
        blockhash = data[-32:]
        if not int.from_bytes(blockhash[:8], 'little'):
            data = data[:-32]
        return {
            'transaction_bytes': data,
            'block_hash': blockhash,
            'txid': txid,
            'transaction_object': Tx.from_bin(data)
        }

    @ldb_batch
    def remove_block(self, blockhash: str):
        key = self.get_key(blockhash, prefix=BLOCK_PREFIX)
        data = self.session.get(self.storage_name + b'.' + key)
        #if data:
        #    txids = utils.split(data[80:], offset=32)
        #    for txid in txids:
        #        self.remove_transaction(txid)
        #else:
        #    Logger.leveldb.warning('remove block on block not found: %s', blockhash)
        key = self.get_key(blockhash, prefix=BLOCK_PREFIX)
        self.session.delete(self.storage_name + b'.' + key)

    @ldb_batch
    def remove_transaction(self, txid: str):
        key = self.get_key(txid, prefix=TRANSACTION_PREFIX)
        self.session.delete(self.storage_name + b'.' + key)



import asyncio
from typing import Dict
import time
from spruned.application.abstracts import HeadersRepository
from spruned.daemon.electrod.electrod_connection import ElectrodConnection
from spruned.daemon.electrod.electrod_interface import ElectrodInterface
from spruned.daemon import exceptions
from spruned.application import database
from spruned.application.logging_factory import Logger
from spruned.application.tools import get_nearest_parent, async_delayed_task


class HeadersReactor:
    """
    This reactor keeps headers aligned to the best height.
    Designed to work with the Electrum Network, it may be ported easily to P2P
    """
    def __init__(
            self,
            repo: HeadersRepository,
            interface: ElectrodInterface,
            loop=asyncio.get_event_loop(),
            store_headers=True,
            delayed_task=async_delayed_task,  # asyncio testing...  :/
            sleep_time_on_inconsistency=20
    ):
        self.repo = repo
        self.interface = interface
        self.loop = loop or asyncio.get_event_loop()
        self.store_headers = store_headers
        self.lock = asyncio.Lock()
        self.subscriptions = []
        self._last_processed_header = None
        self._inconsistencies = []
        self._sync_errors = 0
        self.delayed_task = delayed_task
        self.new_headers_fallback_poll_interval = 60*11
        self.synced = False
        self.sleep_time_on_inconsistency = sleep_time_on_inconsistency
        self.orphans_headers = []
        self.on_best_height_hit_callbacks = []

    def add_on_best_height_hit_callbacks(self, callback):
        self.on_best_height_hit_callbacks.append(callback)

    def set_last_processed_header(self, last):
        if last != self._last_processed_header:
            self._last_processed_header = last
            Logger.electrum.info(
                'Last processed header: %s (%s)',
                self._last_processed_header and self._last_processed_header['block_height'],
                self._last_processed_header and self._last_processed_header['block_hash'],
            )

    async def on_connected(self):
        if self.store_headers:
            self.loop.create_task(self.check_headers())

    async def start(self):
        self.interface.add_header_subscribe_callback(self.on_new_header)
        self.interface.add_on_connected_callback(self.on_connected)
        self.loop.create_task(self.interface.start())

    async def check_headers(self):
        if not self.interface.is_pool_online:
            Logger.electrum.error(
                'Looks like there is no internet connection. check_headers delayed %s',
                self.new_headers_fallback_poll_interval
            )
            self.loop.create_task(self.delayed_task(self.check_headers(), self.new_headers_fallback_poll_interval))
            return
        if self._sync_errors >= 100:
            raise exceptions.SprunedException(
                'Fallback headers check: Too many sync errors. Suspending Sync'
            )
        if not self.synced or self.lock.locked():
            Logger.electrum.debug(
                'Fallback headers check: Not synced yet or sync locked (%s), retrying fallback headers check in 30s',
                self.lock.locked()
            )
            self.loop.create_task(self.delayed_task(self.check_headers(), 30))
            return

        since_last_header = self._last_processed_header and int(time.time()) - self._last_processed_header['timestamp']
        if int(since_last_header) < self.new_headers_fallback_poll_interval:
            retry_in = self.new_headers_fallback_poll_interval - since_last_header
            retry_in = retry_in > 0 and retry_in or self.new_headers_fallback_poll_interval // 2
            Logger.electrum.debug(
                'Fallback headers check: No best header or too recent header (%s), trying again in %s',
                since_last_header, retry_in
            )
            self.loop.create_task(self.delayed_task(self.check_headers(), retry_in))
            return
        Logger.electrum.debug(
            'Fallback headers check, current height: %s, errs: %s',
            self._last_processed_header and self._last_processed_header['block_height'],
            self._sync_errors
        )
        try:
            best_header_response = self._last_processed_header and \
                                   await self.interface.get_header(
                                       self._last_processed_header['block_height'] + 1,
                                       fail_silent_out_of_range=True,
                                       get_peer=True
                                   )
            if best_header_response is None:
                Logger.electrum.debug(
                    'Fallback headers check: Looks like current header (%s) is best header',
                    self._last_processed_header and self._last_processed_header['block_height']
                )
                self.loop.create_task(self.delayed_task(self.check_headers(), self.new_headers_fallback_poll_interval))
                return
            peer, network_best_header = best_header_response
            Logger.electrum.debug('Best header obtained from peer %s: on_new_header(): %s', peer, network_best_header)
        except (
                exceptions.NoQuorumOnResponsesException,
                exceptions.NoPeersException,
                exceptions.NoHeadersException
        ):
            Logger.electrum.warning(
                'Fallback headers check: Electrod is not able to find peers to sync headers. Sleeping 30 secs'
            )
            self.loop.create_task(self.delayed_task(self.check_headers(), 30))
            return

        if network_best_header and network_best_header != self._last_processed_header:
            self.loop.create_task(self.on_new_header(peer, network_best_header))
            self.loop.create_task(self.delayed_task(self.check_headers(), 30))  # loop or fallback
            Logger.electrum.debug('Fallback headers check: Rescheduling sync_header in %ss', 30)
        else:
            self.loop.create_task(self.delayed_task(self.check_headers(), self.new_headers_fallback_poll_interval))
            Logger.electrum.debug(
                'Fallback headers check: Rescheduling sync_headers in %ss',
                self.new_headers_fallback_poll_interval
            )

    async def on_new_header(self, peer, network_best_header: Dict, _r=0):
        if not network_best_header:
            Logger.electrum.warning('Weird. No best header received on call')
            return
        try:
            not _r and await self.lock.acquire()
            if self._last_processed_header and \
                    self._last_processed_header['block_hash'] == network_best_header['block_hash'] and \
                    self._last_processed_header['block_height'] == network_best_header['block_height']:
                self.synced = True
                return
            local_best_header = self.repo.get_best_header()

            if not local_best_header or local_best_header['block_height'] < network_best_header['block_height']:
                self.synced = False
                await self.on_local_headers_behind(local_best_header, network_best_header)
                return
            elif local_best_header['block_height'] > network_best_header['block_height']:
                await self.on_network_headers_behind(network_best_header, peer=peer)
                return

            block_hash = self.repo.get_block_hash(network_best_header['block_height'])
            if block_hash and block_hash != network_best_header['block_hash']:
                await self.interface.handle_peer_error(peer)
                Logger.electrum.error('Inconsistency error with peer %s: (%s), %s',
                                      peer.server_info, network_best_header, block_hash
                                      )
                await asyncio.sleep(self.sleep_time_on_inconsistency)
                if not await self.on_inconsistent_header_received(peer, network_best_header, block_hash):
                    return
            self.set_last_processed_header(network_best_header)
            self.synced = True
        except (
                exceptions.NoQuorumOnResponsesException,
                exceptions.NoPeersException,
                exceptions.NoHeadersException,
        ) as e:
            self._sync_errors += 1
            if _r < 5:
                return await self.on_new_header(peer, network_best_header, _r + 1)
            Logger.electrum.error('Excessive recursion on new_header. %s', e)
        finally:
            if self.synced and self.on_best_height_hit_callbacks:
                while 1:
                    callback = self.on_best_height_hit_callbacks and self.on_best_height_hit_callbacks.pop(0) or None
                    if not callback:
                        break
                    self.loop.create_task(callback)
            not _r and self.lock.release()

    @database.atomic
    async def on_inconsistent_header_received(self, peer: ElectrodConnection, received_header: Dict, local_hash: str):
        """
        received an inconsistent header, this network header differs for hash from
        one at the same height saved in the db.
        check which one we should trust
        """
        response = await self.interface.get_header(received_header['block_height'], fail_silent_out_of_range=True)
        if not response:
            return
        if response['block_hash'] == local_hash:
            Logger.electrum.warning('Received a controversial header (%s), handling error with peer %s',
                                    received_header, peer.server_info)
            await self.interface.handle_peer_error(peer)
            return True

        elif response['block_hash'] == received_header['block_hash']:
            Logger.electrum.error('Remote peers agree the new header is ok, and our is orphan. rolling back')
            orphaned = self.repo.remove_header_at_height(received_header['block_height'])
            await self.on_new_orphan(orphaned)
            self.synced = False
            return

        else:
            Logger.electrum.error(
                'Another inconsistency (net: %s, fetched: %s, local_hash: %s, something must be _very_ wrong',
                received_header, response, local_hash
            )
            self.synced = False
            await self.interface.handle_peer_error(peer)

    @database.atomic
    async def on_new_orphan(self, header: Dict):
        """
        TODO: mark header as orphan
        """
        self.orphans_headers.append(header)
        Logger.electrum.debug('Header %s orphaned, orphans: %s', header['block_hash'], self.orphans_headers)

    @database.atomic
    async def on_local_headers_behind(self, local_best_header: Dict, network_best_header: Dict):
        MAX_SINGLE_HEADERS_BEFORE_USING_CHUNKS = 10

        chunks_at_time = 1
        try:
            if not local_best_header or \
                    local_best_header['block_height'] < network_best_header['block_height'] - \
                    MAX_SINGLE_HEADERS_BEFORE_USING_CHUNKS:
                """
                bootstrap or behind of more than <N> headers
                """
                await self._fetch_headers_chunks(chunks_at_time, local_best_header, network_best_header)
            elif local_best_header['block_height'] == network_best_header['block_height'] - 1:
                """
                behind of 1 header
                """
                await self._save_header(network_best_header)
            else:
                """
                behind of less than MAX_SINGLE_HEADERS_BEFORE_USING_CHUNKS, download single headers and don't use chunks
                """
                await self._fetch_multiple_headers(local_best_header, network_best_header)
        except exceptions.HeadersInconsistencyException:
            Logger.electrum.error('Inconsistency error, rolling back')
            self.set_last_processed_header(None)
            await self.handle_headers_inconsistency()
            return

    async def _fetch_multiple_headers(self, local_best_header: Dict, network_best_header: Dict):
        # The last saved header is old! It must be a while since the last time spruned synced.
        # Download the headers with a chunk, instead of tons of calls to servers..
        headers = await self.interface.get_headers_in_range(
            local_best_header['block_height'],
            network_best_header['block_height'],
        )
        if not headers:
            Logger.electrum.warning('Missing headers on <on_local_headers_behind>')
            await asyncio.sleep(3)
            raise exceptions.NoHeadersException
        saved_headers = self.repo.save_headers(headers[1:])
        self.set_last_processed_header(saved_headers[-1])
        self.synced = True

    async def _save_header(self, network_best_header: Dict):
        # A new header is found, download again from multiple peers to verify it.
        Logger.electrum.debug('Fetching headers')
        self.repo.save_header(
            network_best_header['block_hash'],
            network_best_header['block_height'],
            network_best_header['header_bytes'],
            network_best_header['prev_block_hash']
        )
        self.set_last_processed_header(network_best_header)
        self.synced = True

    async def _fetch_headers_chunks(self, chunks_at_time, local_best_header, network_best_header):
        """
        fetch chunks from local height to network best height, download <chunks_at_time> (>1 is unstable)
        """
        i = 0
        current_height = local_best_header and local_best_header['block_height'] or 0
        while 1:
            Logger.electrum.debug(
                'Behind of %s blocks, fetching chunks',
                network_best_header['block_height'] - current_height

            )
            local_best_height = local_best_header and local_best_header['block_height'] or 0
            rewind_from = get_nearest_parent(local_best_height, 2016) // 2016 + i
            _from = rewind_from
            _to = rewind_from + chunks_at_time
            if _from > (network_best_header['block_height'] // 2016):
                # fixme, move the chunk stuff inside the electrod interface and
                # here just "fetch headers". stop.
                self.synced = True
                return
            res = await self.interface.get_headers_in_range_from_chunks(_from, _to, get_peer=True)
            peer, headers = res if res else (None, None)
            if not headers:
                raise exceptions.NoHeadersException
            saving_headers = [h for h in headers if h['block_height'] > local_best_height] if local_best_height \
                else headers
            try:
                saved_headers = headers and self.repo.save_headers(saving_headers)
            except exceptions.HeadersInconsistencyException:
                await peer.disconnect()
                raise
            Logger.electrum.debug(
                'Fetched %s headers (from chunk %s to chunk %s), saved %s headers of %s',
                len(headers), _from, _to, len(saved_headers), len(saving_headers)
            )
            if saved_headers:
                self.set_last_processed_header(saved_headers[-1])
                current_height = self._last_processed_header['block_height']
            else:
                self.set_last_processed_header(None)
            i += 1

    @database.atomic
    async def on_network_headers_behind(self, network_best_header: Dict, peer=None):
        Logger.electrum.warning('Network headers behind current, closing with peer in 3s')
        await asyncio.sleep(3)
        await self.ensure_consistency(network_best_header, peer)

    async def ensure_consistency(self, network_best_header: Dict, peer: ElectrodConnection):
        repo_header = self.repo.get_header_at_height(network_best_header['block_height'])
        if repo_header['block_hash'] != network_best_header['block_hash']:
            Logger.electrum.error(
                'Warning! A peer (%s), behind in height, '
                'have an header (%s) which differ from our, '
                'saved in the db: %s',
                str(peer.hostname), network_best_header, repo_header
            )
            # FIXME. understand what's going on, maybe rollback
            # TODO - This is STILL an important issue.

        await self.interface.disconnect_from_peer(peer)
        Logger.electrum.info('Closing with peer %s', str(peer.hostname))

    @database.atomic
    async def handle_headers_inconsistency(self):
        self.set_last_processed_header(None)
        local_best_header = self.repo.get_best_header()
        remove_headers_since = get_nearest_parent(local_best_header['block_height'], 2016)
        self.repo.remove_headers_after_height(remove_headers_since)
        Logger.electrum.warning(
            'Headers inconsistency found, removed headers since %s. Current local: %s',
            remove_headers_since, local_best_header['block_height']
        )



import asyncio
import os
import binascii
import time
from typing import Dict
import async_timeout
from connectrum import ElectrumErrorResponse
from connectrum.client import StratumClient
from connectrum.svr_info import ServerInfo
from spruned.application.logging_factory import Logger
from spruned.application.tools import async_delayed_task, check_internet_connection
from spruned.daemon import exceptions
from spruned.daemon.connection_base_impl import BaseConnection
from spruned.daemon.connectionpool_base_impl import BaseConnectionPool


class ElectrodConnection(BaseConnection):
    def __init__(
            self, hostname: str, protocol: str, keepalive=180,
            client=StratumClient, serverinfo=ServerInfo, nickname=None, use_tor=False, loop=None,
            start_score=10, timeout=10, expire_errors_after=180,
            is_online_checker: callable=None, delayer=async_delayed_task
    ):

        self.protocol = protocol
        self.keepalive = keepalive
        self.client = client()
        self.serverinfo_factory = serverinfo
        self.client.keepalive_interval = keepalive
        self.nickname = nickname or binascii.hexlify(os.urandom(8)).decode()
        super().__init__(
            hostname=hostname, use_tor=use_tor, loop=loop, start_score=start_score,
            is_online_checker=is_online_checker, timeout=timeout, delayer=delayer,
            expire_errors_after=expire_errors_after
        )

    @property
    def connected(self):
        return bool(self.client.protocol)

    async def connect(self):
        try:
            with async_timeout.timeout(self._timeout):
                await self.client.connect(
                    self.serverinfo_factory(self.nickname, hostname=self.hostname, ports=self.protocol, version="1.2"),
                    disconnect_callback=self.on_connectrum_disconnect,
                    disable_cert_verify=True,
                    use_tor=self.use_tor
                )
                self._version = self.client.server_version
                Logger.electrum.debug('Connected to %s', self.hostname)
                await self.on_connect()
        except Exception as e:
            Logger.electrum.debug('Exception connecting to %s (%s)', self.hostname, e)
            await self.on_error('connect')

    def on_connectrum_disconnect(self, *_, **__):
        for callback in self._on_disconnect_callbacks:
            self.loop.create_task(callback(self))

    async def ping(self, timeout=2) -> (None, float):
        try:
            async with async_timeout.timeout(timeout):
                now = time.time()
                await self.client.RPC('server.version', 'spruned 1.2', '1.2')
                return time.time() - now
        except asyncio.TimeoutError:
            return

    async def rpc_call(self, method: str, args):
        try:
            async with async_timeout.timeout(self._timeout):
                return await self.client.RPC(method, *args)
        except asyncio.InvalidStateError:
            raise
        except ElectrumErrorResponse as e:
            if e.args and isinstance(e.args[0], dict):
                return e.args[0]
        except Exception as e:
            Logger.electrum.warning(
                'exception on rpc call: %s, %s, %s', self.client.server_info, self.client.protocol, e
            )
            self.loop.create_task(self.delayer(self.on_error(e)))

    async def subscribe(self, channel: str, on_subscription: callable, on_traffic: callable):
        try:
            async with async_timeout.timeout(self._timeout):
                future, q = self.client.subscribe(channel)
            self.subscriptions.append({channel: q})
            header = await future
            self._last_header = header
            on_subscription and self.loop.create_task(on_subscription(self))
            self.loop.create_task(self._poll_queue(q, on_traffic))
        except Exception as e:
            Logger.electrum.error('subscribe %s on %s failed', channel, self.hostname)
            self.loop.create_task(self.delayer(self.on_error(e)))

    async def _poll_queue(self, queue: asyncio.Queue, callback):
        try:
            header = await queue.get()
            Logger.electrum.debug('new data from queue: %s', header)
            self._last_header = header[0]
            self.loop.create_task(self.delayer(callback(self)))
            self.loop.create_task(self.delayer(self._poll_queue(queue, callback)))
        except Exception as e:
            Logger.electrum.error('queue poll failed')
            self.loop.create_task(self.delayer(self.on_error(e)))

    async def disconnect(self):
        try:
            self.client.close()
        except Exception as e:
            Logger.electrum.error('Exception disconnecting from peer: %s', e)
            self.client.protocol = None


class ElectrodConnectionPool(BaseConnectionPool):
    def __init__(
            self,
            connection_factory=ElectrodConnection,
            peers=list(),
            network_checker=check_internet_connection,
            delayer=async_delayed_task,
            loop=asyncio.get_event_loop(),
            use_tor=False,
            connections=3,
            sleep_no_internet=30,
            rpc_call_timeout=30
    ):
        super().__init__(
            peers=peers, network_checker=network_checker, delayer=delayer,
            loop=loop, use_tor=use_tor, connections=connections, sleep_no_internet=sleep_no_internet
        )
        self._connections_keepalive_time = 120
        self._connection_factory = connection_factory
        self.rpc_call_timeout = rpc_call_timeout

    async def connect(self):
        await self._check_internet_connectivity()
        self._keepalive = True
        while self._keepalive:
            if not self.is_online():
                Logger.electrum.error(
                    'Looks like there is no internet connection available. '
                    'Sleeping the connection loop for %s',
                    self._sleep_on_no_internet_connectivity
                )
                await asyncio.sleep(self._sleep_on_no_internet_connectivity)
                await self._check_internet_connectivity()
                continue
            missings = int(self._required_connections - len(self.established_connections))
            if missings > 0:
                Logger.electrum.debug('ConnectionPool: connect, needed: %s', missings)
                self.loop.create_task(self._connect_servers(missings))
            elif missings < 0:
                Logger.electrum.warning('Too many peers.')
                connection = self._pick_connection(fail_silent=True)
                self.loop.create_task(connection.disconnect())
            elif not self._connection_notified:
                for observer in self._on_connect_observers:
                    self.loop.create_task(observer())
                self._connection_notified = True
            await asyncio.sleep(missings and 2 or 10)

    async def _connect_servers(self, howmany: int):
        peers = self._pick_multiple_peers(howmany)
        peers and Logger.electrum.debug('Connecting to peers (%s)', howmany)
        for peer in peers:
            instance = self._connection_factory(
                hostname=peer[0],
                protocol=peer[1],
                keepalive=self._connections_keepalive_time,
                use_tor=self._use_tor,
                loop=self.loop,
                is_online_checker=self.is_online,
                timeout=self.rpc_call_timeout
            )
            instance.add_on_connect_callback(self.on_peer_connected)
            instance.add_on_header_callbacks(self.on_peer_received_header)
            instance.add_on_peers_callback(self.on_peer_received_peers)
            instance.add_on_error_callback(self.on_peer_error)
            self._connections.append(instance)
            Logger.electrum.debug('Created client instance: %s', peer[0])
            self.loop.create_task(instance.connect())

    async def call(self, method, *params, agreement=1, get_peer=False, fail_silent=False) -> (None, Dict):
        if get_peer and agreement > 1:
            raise ValueError('Error!')
        if agreement > self._required_connections:
            raise ValueError('Agreement requested is out of range %s' % self._required_connections)
        if agreement > len(self.established_connections):
            raise exceptions.NoPeersException
        if agreement > 1:
            connections = self._pick_multiple_connections(agreement)
            responses = await asyncio.gather(
                *[connection.rpc_call(method, params) for connection in connections]
            )
            responses = [r for r in responses if r is not None]
            if len(responses) < agreement:
                Logger.electrum.exception('call, requested %s responses, received %s', agreement, len(responses))
                Logger.electrum.debug('call, requested %s responses, received %s', agreement, responses)
                if fail_silent:
                    return
                raise exceptions.ElectrodMissingResponseException
            try:
                return self._handle_responses(responses)
            except exceptions.NoQuorumOnResponsesException:
                if fail_silent:
                    return
                raise
        connection = self._pick_connection()
        response = await connection.rpc_call(method, params)
        if not response and not fail_silent:
            await self.on_peer_error(connection)
            raise exceptions.ElectrodMissingResponseException(connection)
        return (connection, response) if get_peer else response

    @staticmethod
    def _handle_responses(responses) -> Dict:
        if len(responses) == 1:
            return responses and responses[0]
        for response in responses:
            if responses.count(response) == len(responses):
                return response
        raise exceptions.NoQuorumOnResponsesException(responses)

    def on_peer_received_peers(self, peer: ElectrodConnection, *_):
        raise NotImplementedError

    async def on_peer_connected(self, peer: ElectrodConnection):
        future = peer.subscribe(
            'blockchain.headers.subscribe',
            self.on_peer_received_header,
            self.on_peer_received_header
        )
        self.loop.create_task(self.delayer(future))



import plyvel
from sqlalchemy import Column, String, Integer, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
import threading
from functools import wraps
from spruned import settings

Base = declarative_base()


class Header(Base):
    __tablename__ = 'headers'
    id = Column(Integer, primary_key=True)
    blockheight = Column(Integer, index=True, unique=True)
    blockhash = Column(String, index=True, unique=True)
    data = Column(String)


engine = create_engine('sqlite:///' + settings.SQLITE_DBNAME)
Base.metadata.create_all(engine)

sqlite = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=engine))

if not settings.TESTING:
    storage_ldb = plyvel.DB(settings.LEVELDB_BLOCKCHAIN_ADDRESS, create_if_missing=True)
else:
    from unittest.mock import Mock
    storage_ldb = Mock()

_local = threading.local()
_local.session = sqlite
_local.storage_ldb = storage_ldb


def atomic(fun):
    @wraps(fun)
    def decorator(*args, **kwargs):
        try:
            try:
                _local.counter += 1
            except AttributeError:
                _local.counter = 1
            r = fun(*args, **kwargs)
            if _local.counter == 1:
                _local.session.commit()
                _local.counter -= 1
            return r
        except Exception as e:
            if _local.counter == 1:
                _local.session.rollback()
                _local.counter -= 1
            raise e
        finally:
            _local.session.close()
    return decorator


def ldb_batch(fun):
    @wraps(fun)
    def decorator(*args, **kwargs):
        try:
            _local.leveldb_counter += 1
        except AttributeError:
            _local.leveldb_counter = 1
        if _local.leveldb_counter == 1:
            try:
                if not _local.in_ldb_batch:
                    _local.storage_ldb = storage_ldb.write_batch()
                    _local.in_ldb_batch = True
            except AttributeError:
                _local.in_ldb_batch = True
                _local.storage_ldb = storage_ldb.write_batch()
        r = fun(*args, **kwargs)
        if _local.leveldb_counter == 1:
            _local.storage_ldb.write()
            if _local.in_ldb_batch:
                _local.in_ldb_batch = False
                del _local.storage_ldb
                _local.storage_ldb = storage_ldb
        _local.leveldb_counter -= 1
        return r
    return decorator



import asyncio
from unittest import TestCase
from unittest.mock import Mock, call
from spruned.daemon.bitcoin_p2p import P2PConnectionPool
from test.utils import async_coro, coro_call


class TestP2PConnectionPool(TestCase):
    def setUp(self):
        self.loopmock = Mock()
        self.online_checker = Mock()
        self.delayer = Mock()
        self.port = 8333
        self.sut = P2PConnectionPool(
            network_checker=self.online_checker, delayer=self.delayer,
            loop=self.loopmock, use_tor=False, sleep_no_internet=1, connections=2
        )
        self.loop = asyncio.get_event_loop()
        self.peers = [['cafe', 123], ['babe', 123]]

    def test_connect_ok(self):
        async def parallel():
            await asyncio.sleep(3)
            for peer in self.peers:
                self.sut.add_peer(peer)

            await asyncio.sleep(5)
            self.sut._keepalive = False

        self.online_checker.side_effect = [async_coro(False), async_coro(True)]
        self.loop.run_until_complete(
            asyncio.gather(
                parallel(),
                self.sut.connect()
            )
        )
        Mock.assert_has_calls(
            self.loopmock.create_task,
            calls=[
                call(coro_call('_connect_peer')), call(coro_call('_connect_peer'))
            ]
        )



import asyncio
import random
from unittest import TestCase
from unittest.mock import Mock, call
from spruned.application.jsonrpc_server import JSONRPCServer
from spruned.application.utils.jsonrpc_client import JSONClient
from spruned.daemon import exceptions
from test.utils import async_coro


class TestJSONRPCServerGetrawtransaction(TestCase):
    def setUp(self):
        bindport = random.randint(31337, 41337)
        self.sut = JSONRPCServer('127.0.0.1', bindport, 'testuser', 'testpassword')
        self.vo_service = Mock()
        self.sut.set_vo_service(self.vo_service)
        self.client = JSONClient(b'testuser', b'testpassword', '127.0.0.1', bindport)
        self.loop = asyncio.get_event_loop()

    def test_getrawtransaction_success(self):
        self.vo_service.getrawtransaction.side_effect = [async_coro('cafebabe')]

        async def test():
            await self.sut.start()
            response = await self.client.call('getrawtransaction', params=['00' * 32])
            response2 = await self.client.call('getrawtransaction', params=['00' * 32, True])
            return response, response2

        res, res2 = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': None, 'id': 1, 'jsonrpc': '2.0', 'result': 'cafebabe'}
        )
        self.assertEqual(
            res2,
            {
                'error': {'code': -8, 'message': 'spruned error: Verbose mode not supported'},
                'id': 1,
                'jsonrpc': '2.0',
                'result': None
            }
        )
        Mock.assert_has_calls(
            self.vo_service.getrawtransaction,
            calls=[
                call('00' * 32, False)
            ]
        )

    def test_getrawtransaction_error_missing(self):
        self.vo_service.getrawtransaction.side_effect = [async_coro(None)]

        async def test():
            await self.sut.start()
            response = await self.client.call('getrawtransaction', params=['00' * 32])
            return response

        res = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {
                'error': {'code': -5,
                          'message': 'No such mempool or blockchain transaction. [maybe try again]'
                          },
                'id': 1,
                'jsonrpc': '2.0',
                'result': None
            }
        )

    def test_getrawtransaction_error_params(self):
        response = None
        self.vo_service.getrawtransaction.return_value = async_coro(response)

        async def test():
            await self.sut.start()
            response1 = await self.client.call('getrawtransaction', params=['wrong_hash'])
            response2 = await self.client.call('getrawtransaction', params=['cafebabe'])
            return response1, response2

        res, res2 = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': {'code': -8,
                       'message': 'parameter 1 must be hexadecimal string (not '
                                  "'wrong_hash')"},
             'id': 1,
             'jsonrpc': '2.0',
             'result': None}
        )
        self.assertEqual(
            res2,
            {'error': {'code': -8, 'message': "parameter 1 must be of length 64 (not '8')"},
             'id': 1,
             'jsonrpc': '2.0',
             'result': None}
        )
        Mock.assert_not_called(self.vo_service.getblockheader)

    def test_getrawtransaction_error_genesis(self):
        self.vo_service.getrawtransaction.side_effect = exceptions.GenesisTransactionRequestedException

        async def test():
            await self.sut.start()
            response1 = await self.client.call(
                'getrawtransaction',
                params=[
                    '000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f'
                ])
            return response1

        res = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': {'code': -5,
                       'message': 'The genesis block coinbase is not considered an '
                                  'ordinary transaction and cannot be retrieved'
                       },
             'id': 1,
             'jsonrpc': '2.0',
             'result': None}
        )




import asyncio

from spruned import settings
from spruned.application.database import ldb_batch
from spruned.application.logging_factory import Logger
from spruned.repositories.headers_repository import HeadersSQLiteRepository
from spruned.repositories.blockchain_repository import BlockchainRepository, BLOCK_PREFIX, TRANSACTION_PREFIX


class Repository:
    def __init__(self, headers, blocks, keep_blocks=200):
        self._headers_repository = headers
        self._blockchain_repository = blocks
        self.ldb = None
        self.sqlite = None
        self.cache = None
        self.keep_blocks = keep_blocks
        self.integrity_lock = asyncio.Lock()

    @property
    def headers(self) -> HeadersSQLiteRepository:
        return self._headers_repository

    @property
    def blockchain(self) -> BlockchainRepository:
        return self._blockchain_repository

    @classmethod
    def instance(cls):  # pragma: no cover
        from spruned.application import database
        from spruned.application.context import ctx
        headers_repository = HeadersSQLiteRepository(database.sqlite)
        blocks_repository = BlockchainRepository(
            database.storage_ldb,
            settings.LEVELDB_BLOCKCHAIN_SLUG,
            settings.LEVELDB_BLOCKCHAIN_ADDRESS
        )
        i = cls(
            headers=headers_repository,
            blocks=blocks_repository,
            keep_blocks=ctx.keep_blocks
        )
        i.sqlite = database.sqlite
        i.ldb = database.storage_ldb
        return i

    async def ensure_integrity(self):
        try:
            await self.integrity_lock.acquire()
            self._ensure_no_stales_in_blockchain_repository()
        finally:
            self.integrity_lock.release()

    @ldb_batch
    def _ensure_no_stales_in_blockchain_repository(self):
        Logger.leveldb.debug('Ensuring no stales in blockchain repository')
        keypref = self.blockchain.storage_name + b'.' + BLOCK_PREFIX
        extemp = self.get_extemped_blockhash()
        keep_keys = [self.blockchain.get_key(e, keypref) for e in extemp]
        index = self.cache.get_index()
        if not index:
            Logger.cache.debug('Cache index not found')
            return
        index = [self.blockchain.storage_name + b'.' + k for k in index.get('keys', {}).keys()]
        if not index:
            Logger.cache.debug('Empty index found')
        iterator = self.ldb.iterator()
        purged = 0
        txs = 0
        cached = 0
        tot = -1  # skip cache index
        kept = 0
        for x in iterator:
            tot += 1
            if keypref not in x[0]:
                if x[0] == self.cache.cache_name:
                    continue
                elif self.blockchain.storage_name + b'.' + TRANSACTION_PREFIX in x[0]:
                    txs += 1
                    continue
            if x[0] in keep_keys:
                kept += 1
                continue
            elif x[0] in index:
                cached += 1
                continue
            elif x[0] not in index:
                self.blockchain.remove_block(x[0].replace(keypref + b'.', b''))
                purged += 1
            else:
                raise ValueError(x)
        Logger.cache.info(
            '\nPurged from storage %s elements not tracked by cache.\n'
            'Total tracked: %s\n'
            'Total protected: %s,\n'
            'Total cached: %s,\n'
            'Total entries: %s,\n'
            'Total transactions: %s\n',
            purged, len(index), kept, cached, tot, txs
        )
        return

    def set_cache(self, cache):
        self.cache = cache
        self.headers.set_cache(cache)
        self.blockchain.set_cache(cache)

    def get_extemped_blockhash(self):
        best_header = self.headers.get_best_header()
        _keep_to = best_header and best_header.get('block_height') or 0
        keep_from = _keep_to - self.keep_blocks if _keep_to - self.keep_blocks > 0 else 0
        keep_headers = keep_from and self.headers.get_headers_since_height(keep_from) or []
        keep_hashes = [k.get('block_hash') for k in keep_headers]
        return keep_hashes



import asyncio
from spruned.daemon import exceptions
from spruned.application.logging_factory import Logger
from spruned.application.tools import async_delayed_task
from spruned.daemon.bitcoin_p2p.p2p_interface import P2PInterface
from spruned.repositories.repository import Repository


class BlocksReactor:
    """
    This reactor keeps non-pruned blocks aligned to the best height.
    """
    def __init__(
            self,
            repository: Repository,
            interface: P2PInterface,
            loop=asyncio.get_event_loop(),
            prune=200,
            delayed_task=async_delayed_task
    ):
        self.repo = repository
        self.interface = interface
        self.loop = loop or asyncio.get_event_loop()
        self.lock = asyncio.Lock()
        self.delayer = delayed_task
        self._last_processed_block = None
        self._prune = prune
        self._max_per_batch = 10
        self._available = False
        self._fallback_check_interval = 30

    def set_last_processed_block(self, last):
        if last != self._last_processed_block:
            self._last_processed_block = last
            Logger.p2p.info(
                'Last processed block: %s (%s)',
                self._last_processed_block and self._last_processed_block['block_height'],
                self._last_processed_block and self._last_processed_block['block_hash'],
            )

    def on_header(self, best_header):
        Logger.p2p.debug('BlocksReactor.on_header: %s', best_header)
        self.loop.create_task(self._check_blockchain(best_header))

    async def check(self):
        urgent = False
        try:
            best_header = self.repo.headers.get_best_header()
            urgent = await self._check_blockchain(best_header)
        except Exception as e:
            urgent = urgent or False
            Logger.p2p.exception('Error on BlocksReactor fallback %s', str(e))
        finally:
            self.loop.create_task(
                self.delayer(self.check, 0 if urgent else self._fallback_check_interval)
            )

    async def _check_blockchain(self, best_header):
        urgent = False
        try:
            await self.lock.acquire()
            if not self._last_processed_block or \
                    best_header['block_height'] > self._last_processed_block['block_height']:
                urgent = await self._on_blocks_behind_headers(best_header)
            elif best_header['block_height'] < self._last_processed_block['block_height']:
                Logger.p2p.warning('Headers index is behind what this task done. Reset current status')
                self.set_last_processed_block(None)
                urgent = True
                # This will be fixed in the next iteration by on_blocks_behind_header
            else:
                if best_header['block_hash'] != self._last_processed_block['block_hash']:
                    Logger.p2p.warning('There must be a reorg. Reset current status')
                    # This will be fixed in the next iteration by on_blocks_behind_header
                    self.set_last_processed_block(None)
                    urgent = True
        except (
            exceptions.BlocksInconsistencyException
        ):
            Logger.p2p.exception('Exception checkping the blockchain')
            self.set_last_processed_block(None)
            urgent = True
        finally:
            self.lock.release()
            return urgent

    async def _on_blocks_behind_headers(self, best_header):
        if self._last_processed_block and \
                best_header['block_height'] - self._last_processed_block['block_height'] < self._prune:
            height_to_start = self._last_processed_block['block_height']
            urgent = False
        else:
            height_to_start = best_header['block_height'] - self._prune
            height_to_start = height_to_start if height_to_start >= 0 else 0
            urgent = True

        headers = self.repo.headers.get_headers_since_height(height_to_start, limit=self._max_per_batch)
        _local_blocks = {h['block_hash']: self.repo.blockchain.get_block(h['block_hash'],
                                                                         with_transactions=False) for h in headers}
        _local_hblocks = {k: v for k, v in _local_blocks.items() if v is not None}
        _request = [x['block_hash'] for x in headers if x['block_hash'] not in _local_hblocks]
        blocks = _request and await self.interface.get_blocks(*_request)
        _hheaders = {v['block_hash']: v for v in headers}
        if blocks:
            urgent = urgent or False
            try:
                sorted_values = sorted(blocks.values(), key=lambda x: x['block_hash'])
                saved_blocks = self.repo.blockchain.save_blocks(*sorted_values)
                Logger.p2p.debug('Saved block %s', saved_blocks)
            except:
                Logger.p2p.exception('Error saving blocks %s', blocks)
                return True
        else:
            urgent = True
            saved_blocks = [_local_hblocks[headers[-1]['block_hash']]]

        if saved_blocks:
            self.set_last_processed_block(
                {
                    'block_hash': saved_blocks[-1]['block_hash'],
                    'block_height': _hheaders[saved_blocks[-1]['block_hash']]['block_height']
                }
            )
        else:
            urgent = True
        return urgent

    async def on_connected(self):
        self._available = True
        self.loop.create_task(self.check())

    async def start(self):
        self.interface.add_on_connect_callback(self.on_connected)
        self.loop.create_task(self.interface.start())

    async def bootstrap_blocks(self):
        while len(self.interface.pool.established_connections) < self.interface.pool.required_connections:
            Logger.p2p.info('Bootstrap: ConnectionPool not ready yet')
            await asyncio.sleep(5)
        Logger.p2p.info('Bootstrap: Starting Bootstrap Procedure on %s blocks', self._prune)
        try:
            await self.lock.acquire()
            best_header = self.repo.headers.get_best_header()
            headers = self.repo.headers.get_headers_since_height(best_header['block_height'] - self._prune)
            missing_blocks = []
            for blockheader in headers:
                if not self.repo.blockchain.get_block(blockheader['block_hash']):
                    missing_blocks.append(blockheader['block_hash'])
            i = 0
            while 1:
                i += 1
                if len(self.interface.pool.established_connections) - len(self.interface.pool._busy_peers) \
                        < self.interface.pool.required_connections:
                    Logger.p2p.debug('Missing peers. Waiting.')
                    await asyncio.sleep(20)
                    continue

                status = float(100) / self._prune * (len(headers) - len(missing_blocks))
                status = status if status <= 100 else 100
                self.interface.set_bootstrap_status(status)
                missing_blocks = missing_blocks[::-1]
                _blocks = [
                    missing_blocks.pop() for _ in
                    range(0, int(len(self.interface.pool.established_connections)*0.5) or 1)
                    if missing_blocks
                ]
                if not _blocks:
                    Logger.p2p.info('Bootstrap: No blocks to fetch.')
                    break
                not i and Logger.p2p.info('Bootstrap: Fetching %s blocks', len(_blocks))

                async def save_block(blockhash):
                    block = (await asyncio.gather(
                        self.interface.get_block(blockhash, peers=1, timeout=20),
                        return_exceptions=True
                    ))[0]
                    if isinstance(block, dict):
                        Logger.p2p.info(
                            'Bootstrap: saved block %s (%s/%s)',
                            block['block_hash'],
                            self._prune - len(missing_blocks),
                            self._prune
                        )
                        self.repo.blockchain.save_block(block)
                    else:
                        Logger.p2p.debug('Bootstrap: enqueuing block %s (%s)', blockhash, type(block))
                        missing_blocks.insert(0, blockhash)

                futures = [save_block(blockhash) for blockhash in _blocks]
                await asyncio.gather(*futures, return_exceptions=True)
        finally:
            self.lock.release()



import asyncio
import async_timeout
import time

from pycoin.message import InvItem
from pycoin.message.InvItem import ITEM_TYPE_TX
from pycoinnet.Peer import Peer
from pycoinnet.PeerEvent import PeerEvent
from pycoinnet.networks import MAINNET
from pycoinnet.inv_batcher import InvBatcher
from pycoinnet.version import version_data_for_peer, NODE_WITNESS, NODE_NONE
from spruned.application.logging_factory import Logger
from spruned.application.tools import check_internet_connection, async_delayed_task
from spruned.daemon.connection_base_impl import BaseConnection
from spruned.daemon.connectionpool_base_impl import BaseConnectionPool


class P2PConnection(BaseConnection):
    def ping(self, timeout=None):
        self.peer.send_msg('ping', int(time.time()))

    def __init__(
            self, hostname, port, peer=Peer, network=MAINNET, loop=asyncio.get_event_loop(),
            use_tor=None, start_score=3,
            is_online_checker: callable=None,
            timeout=10, delayer=async_delayed_task, expire_errors_after=180,
            call_timeout=30, connector=asyncio.open_connection):

        super().__init__(
            hostname=hostname, use_tor=use_tor, loop=loop, start_score=start_score,
            is_online_checker=is_online_checker, timeout=timeout, delayer=delayer,
            expire_errors_after=expire_errors_after
        )
        self.port = port
        self._peer_factory = peer
        print('NEW PEER: Network: ', network)
        self._peer_network = network
        self.peer = None
        self._version = None
        self.loop = loop
        self._event_handler = None
        self._call_timeout = call_timeout
        self._on_block_callbacks = []
        self._on_transaction_callbacks = []
        self.connector = connector

    @property
    def connected(self):
        return bool(self.peer)

    def add_on_blocks_callback(self, callback):
        self._on_block_callbacks.append(callback)

    def add_on_transaction_callback(self, callback):
        self._on_transaction_callbacks.append(callback)

    @property
    def peer_event_handler(self) -> PeerEvent:
        return self._event_handler

    async def connect(self):
        try:
            async with async_timeout.timeout(self._timeout):
                reader, writer = await self.connector(host=self.hostname, port=self.port)
                peer = self._peer_factory(
                    reader,
                    writer,
                    self._peer_network.magic_header,
                    self._peer_network.parse_from_data,
                    self._peer_network.pack_from_data
                )
                version_data = version_data_for_peer(
                    peer, version=70015, local_services=NODE_NONE, remote_services=NODE_WITNESS
                )
                peer.version = await peer.perform_handshake(**version_data)
                self._event_handler = PeerEvent(peer)
                self._version = peer.version
                Logger.p2p.info('Connected to peer %s', self.version)
                self.peer = peer
                self._setup_events_handler()
        except Exception as e:
            self.peer = None
            Logger.p2p.debug('Exception connecting to %s (%s)', self.hostname, e)
            self.loop.create_task(self.on_error('connect'))
            return

        self.loop.create_task(self.on_connect())
        return self

    async def on_connect(self):
        for callback in self._on_connect_callbacks:
            self.loop.create_task(callback(self))

    async def disconnect(self):
        try:
            self.peer and self.peer.close()
        except:
            Logger.p2p.error('Error closing with peer: %s', self.peer.peername())
        finally:
            self.peer = None

    def _setup_events_handler(self):
        self.peer_event_handler.set_request_callback('inv', self._on_inv)
        self.peer_event_handler.set_request_callback('addr', self._on_addr)
        self.peer_event_handler.set_request_callback('alert', self._on_alert)
        self.peer_event_handler.set_request_callback('ping', self._on_ping)
        self.peer_event_handler.set_request_callback('sendheaders', self._dummy_handler)
        self.peer_event_handler.set_request_callback('feefilter', self._dummy_handler)
        self.peer_event_handler.set_request_callback('sendcmpct', self._dummy_handler)

    def _dummy_handler(self, *a, **kw):
        pass

    def _on_inv(self, event_handler, name, data):
        try:
            self.loop.create_task(self._process_inv(event_handler, name, data))
        except:
            Logger.p2p.exception('Exception on inv')

    def _on_alert(self, event_handler, name, data):  # pragma: no cover
        try:
            Logger.p2p.debug('Handle alert: %s, %s, %s', event_handler, name, data)
        except:
            Logger.p2p.exception('Exception on alert')

    def _on_addr(self, event_handler, name, data):  # pragma: no cover
        try:
            Logger.p2p.debug('Handle addr: %s, %s, %s', event_handler, name, data)
        except:
            Logger.p2p.exception('Exception on addr')

    def _on_ping(self, event_handler, name, data):
        try:
            self.peer.send_msg("pong", nonce=data["nonce"])
            Logger.p2p.debug('Handle ping: %s, %s, %s', event_handler, name, data)
        except:
            Logger.p2p.exception('Exception on ping')

    async def _process_inv(self, event_handler, name, data):
        txs = 0
        for item in data.get('items'):
            if item.item_type == ITEM_TYPE_TX:
                txs += 1
                for callback in self._on_transaction_callbacks:
                    self.loop.create_task(callback(self, item))
            else:
                Logger.p2p.debug('Unhandled InvType: %s, %s, %s', event_handler, name, item)
        Logger.p2p.debug('Received %s items, txs: %s', len(data.get('items')), txs)


class P2PConnectionPool(BaseConnectionPool):
    async def on_peer_received_peers(self, peer, *a):
        Logger.p2p.debug('Received peers from peer: %s: (%s)', peer.hostname, a)

    def __init__(
            self,
            peers=list(),
            network_checker=check_internet_connection,
            delayer=async_delayed_task,
            loop=asyncio.get_event_loop(),
            use_tor=False,
            connections=3,
            sleep_no_internet=30,
            batcher=InvBatcher,
            network=MAINNET,
            batcher_timeout=30,
            ipv6=False
    ):
        super().__init__(
            peers=peers, network_checker=network_checker, delayer=delayer, ipv6=ipv6,
            loop=loop, use_tor=use_tor, connections=connections, sleep_no_internet=sleep_no_internet
        )
        self._batcher_factory = batcher
        self._network = network
        self._batcher_timeout = batcher_timeout
        self._busy_peers = set()

    @property
    def required_connections(self):
        return self._required_connections

    @property
    def available(self):
        return len(self.connections) >= self._required_connections

    def add_peer(self, peer):
        self._peers.append(peer)

    @property
    def connections(self):
        lost = [connection for connection in self._connections if not connection.peer]
        for l in lost:
            del l
        return self._connections

    async def connect(self):
        await self._check_internet_connectivity()
        self._keepalive = True
        while not self._peers:
            Logger.p2p.warning('Peers not loaded. Sleeping.')
            await asyncio.sleep(5)

        while self._keepalive:
            if not self.is_online():
                Logger.p2p.error(
                    'Looks like there is no internet connection available. '
                    'Sleeping the connection loop for %s',
                    self._sleep_on_no_internet_connectivity
                )
                await asyncio.sleep(self._sleep_on_no_internet_connectivity)
                await self._check_internet_connectivity()
                continue
            missings = self._required_connections - len(self.established_connections)
            if missings > 0:
                peers = self._pick_multiple_peers(missings)
                for peer in peers:
                    host, port = peer
                    self.loop.create_task(self._connect_peer(host, port))
            elif len(self.established_connections) > self._required_connections:
                Logger.p2p.warning('Too many connections')
                connection = self._pick_connection()
                self.loop.create_task(connection.disconnect())
            Logger.p2p.debug(
                'P2PConnectionPool: Sleeping %ss, connected to %s peers', 10, len(self.established_connections)
            )
            for connection in self._connections:
                if connection.score <= 0:
                    self.loop.create_task(self._disconnect_peer(connection))
            await asyncio.sleep(5)

    async def _disconnect_peer(self, peer):
        await peer.disconnect()

    async def _connect_peer(self, host: str, port: int):
        Logger.p2p.debug('Allocating peer %s:%s', host, port)
        connection = P2PConnection(host, port, loop=self.loop, network=self._network)
        if not await connection.connect():
            Logger.p2p.debug(
                'Connection to %s - %s failed. Connected to %s peers', host, port, len(self.established_connections)
            )
            return
        self._connections.append(connection)
        connection.add_on_connect_callback(self.on_peer_connected)
        connection.add_on_header_callbacks(self.on_peer_received_header)
        connection.add_on_peers_callback(self.on_peer_received_peers)
        connection.add_on_error_callback(self.on_peer_error)

    async def get(self, inv_item: InvItem, peers=None, timeout=None):
        batcher = self._batcher_factory()
        connections = []
        s = time.time()
        Logger.p2p.debug('Fetching InvItem %s', inv_item)
        future = None
        try:
            async with async_timeout.timeout(timeout if timeout is not None else self._batcher_timeout):
                connections = self._pick_multiple_connections(peers if peers is not None else 1)
                _ = [self._busy_peers.add(connection.hostname) for connection in connections]
                for connection in connections:
                    Logger.p2p.debug('Adding connection %s to batcher', connection.hostname)
                    await batcher.add_peer(connection.peer_event_handler)
                future = await batcher.inv_item_to_future(inv_item)
                response = await future
                Logger.p2p.debug('InvItem %s fetched in %ss', inv_item, round(time.time() - s, 4))
                return response and response
        except Exception as error:
            Logger.p2p.debug(
                'Error in get InvItem %s, error: %s, failed in %ss', inv_item, str(error), round(time.time() - s, 4)
            )
            for connection in connections:
                connection.add_error()
            future and future.cancel()
        finally:
            try:
                _ = [self._busy_peers.remove(connection.hostname) for connection in connections]
                del connections
            except KeyError as e:
                Logger.p2p.debug('Peer %s already removed from busy peers', str(e))

            def del_batcher(_b):
                try:
                    _b.stop()
                    del _b._inv_item_future_queue
                    del _b._inv_item_hash_to_future[str(inv_item)]
                except:
                    del _b

            self.loop.run_in_executor(None, lambda: del_batcher(batcher))

    async def on_peer_connected(self, peer):
        Logger.p2p.debug('on_peer_connected: %s', peer.hostname)



import os
import subprocess
from setuptools import setup, find_packages
import sys
from setuptools.command.install import install

sys.path.insert(0, '.')
import spruned


def read_pip_requirements():
    reqs = []
    with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), 'r') as f:
        lines = f.readlines()
    for line in lines:
        line = line.strip()
        if '://' not in line:
            reqs.append(line)
    return reqs


def read_dependency_links():
    links = []
    with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), 'r') as f:
        lines = f.readlines()
    for l in lines:
        line = l.strip()
        if '://' in line:
            links.append(line)
    return links


def read_file(name):
    with open(name, 'r', encoding='utf-8') as f:
        file = f.read()
    return file


class InstallGithubDepedenciesCommand(install):
    description = 'install github deps'

    def run(self):
        def install_github_dependencies():

            for dep in read_dependency_links():
                subprocess.call(['pip', 'install', '-e', dep])

        install_github_dependencies()
        install.run(self)


github_links = '\n- '.join(read_dependency_links())
github_libraries = '\n- '.join([x.split('#egg=')[1] for x in read_dependency_links()])
print('\nWarning! Know what you do!')
print('\nSpruned installer will also install custom dependencies from the following github links:\n\n-', github_links)
print('\nIs warmly advised to run spruned into a virtual environment, '
      'especially if you have installed the following libraries:\n\n-', github_libraries)
# Ouch, pip captured this. Find a way to inform the user on what's going on.

setup(
    cmdclass={
        'install': InstallGithubDepedenciesCommand,
    },
    name='spruned',
    version=spruned.__version__,
    url='https://github.com/gdassori/spruned/',
    license='MIT',
    author='Guido Dassori',
    author_email='guido.dassori@gmail.com',
    python_requires='>=3.5.2',
    description='Bitcoin Lightweight Pseudonode',
    long_description=read_file('README.rst'),
    install_requires=read_pip_requirements(),
    packages=find_packages(exclude=['htmlcov']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'spruned = spruned.app:main'
        ]
    },
    include_package_data=True
)



import asyncio
from unittest import TestCase
from unittest.mock import Mock, call, ANY

from pycoinnet.networks import MAINNET

from spruned.daemon.bitcoin_p2p.p2p_interface import P2PInterface
from test.utils import async_coro


class TestP2PInterface(TestCase):
    def setUp(self):
        self.pool = Mock()
        self.loopmock = Mock()
        self.peers_bootstrapper = Mock()
        self.sut = P2PInterface(self.pool, loop=self.loopmock, peers_bootstrapper=self.peers_bootstrapper)
        self.loop = asyncio.get_event_loop()

    def test_start(self):
        self.peers_bootstrapper.return_value = async_coro((('1.2.3.4', 8333), ('1.2.3.5', 8333)))

        self.loop.run_until_complete(self.sut.start())
        self.assertEqual(self.pool.add_on_connected_observer.call_count, 1)
        Mock.assert_called_with(self.peers_bootstrapper, MAINNET)
        Mock.assert_has_calls(
            self.pool.add_peer,
            calls=[
                call(('1.2.3.4', 8333)),
                call(('1.2.3.5', 8333))
            ]
        )
        Mock.assert_called_once_with(self.pool.connect)
        self.assertEqual(self.loopmock.create_task.call_count, 1)
        self.sut.set_bootstrap_status(10)
        self.assertEqual(self.sut.bootstrap_status, 10)

    def test_get_block(self):
        block = Mock(previous_block_hash='ff'*32, timestamp=3)
        block.hash.return_value = 'aa'*32
        block.as_blockheader.return_value = block
        block.as_bin.return_value = b'block'
        self.pool.get.return_value = async_coro(block)
        response = self.loop.run_until_complete(self.sut.get_block('aa'*32))
        self.assertEqual(
            response,
            {
                "block_hash": "aa"*32,
                "prev_block_hash": "ff"*32,
                "timestamp": 3,
                "block_object": ANY,
                "header_bytes": b"block",
                "block_bytes": b"block"
            }
        )

    def test_get_blocks(self):
        block = Mock(previous_block_hash='ff'*32, timestamp=3)
        block.hash.return_value = 'aa'*32
        block.as_blockheader.return_value = block
        block.as_bin.return_value = b'block'
        self.pool.get.return_value = async_coro(block)
        response = self.loop.run_until_complete(self.sut.get_blocks('aa'*32))
        self.assertEqual(
            {
                "aa" * 32:
                    {
                        "block_hash": "aa"*32,
                        "prev_block_hash": "ff"*32,
                        "timestamp": 3,
                        "block_object": ANY,
                        "header_bytes": b"block",
                        "block_bytes": b"block"
                    }
            },
            response
        )
